@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2020-2021 -- leonerd@leonerd.org.uk
 
-package Object::Pad::MOP::Class 0.55;
+package Object::Pad::MOP::Class 0.57;
 
 use v5.14;
 use warnings;
@@ -132,23 +132,39 @@ Returns a list of superclasses, as L<Object::Pad::MOP::Class> instances.
 Because C<Object::Pad> does not support multiple superclasses, this list will
 contain at most one item.
 
-=head2 roles
+=head2 direct_roles
 
-   @roles = $metaclass->roles
+   @roles = $metaclass->direct_roles
 
-Returns a list of roles implemented by this class, as
+Returns a list of the roles introduced by this class (i.e. added by `does`
+declarations but not inherited from the superclass), as
 L<Object::Pad::MOP::Class> instances.
 
-=head2 compose_role
+This method is also aliased as C<roles>.
 
-   $metaclass->compose_role( $rolename )
-   $metaclass->compose_role( $rolemeta )
+=head2 all_roles
+
+   @roles = $metaclass->all_roles
+
+I<Since version 0.56.>
+
+Returns a list of all the roles implemented by this class (i.e. including
+those inherited from the superclass), as L<Object::Pad::MOP::Class> instances.
+
+=head2 add_role
+
+   $metaclass->add_role( $rolename )
+   $metaclass->add_role( $rolemeta )
+
+I<Since verison 0.56.>
 
 Adds a new role to the list of those implemented by the class.
 
 The new role can be specified either as a plain string giving its name, or as
 an C<Object::Pad::MOP::Class> meta instance directly.
 
+Before version 0.56 this was called C<compose_role>.
+
 =head2 add_BUILD
 
    $metaclass->add_BUILD( $code )
@@ -163,9 +179,9 @@ Adds a new named method to the class under the given name, as CODE reference.
 
 Returns an instance of L<Object::Pad::MOP::Method> to represent it.
 
-=head2 get_own_method
+=head2 get_direct_method
 
-   $metamethod = $metaclass->get_own_method( $name )
+   $metamethod = $metaclass->get_direct_method( $name )
 
 Returns an instance of L<Object::Pad::MOP::Method> to represent the method of
 the given name, if one exists. If not an exception is thrown.
@@ -175,6 +191,40 @@ C<method> keyword on the class itself, or added via L</add_method>. This will
 not see other names in the package stash, even if they contain a C<CODE> slot,
 nor will it see methods inherited from a superclass.
 
+This is also aliased as C<get_own_method> for compatibility with the
+L<MOP::Class> interface.
+
+=head2 get_method
+
+   $metamethod = $metaclass->get_method( $name )
+
+I<Since version 0.57.>
+
+Returns an instance of L<Object::Pad::MOP::Method> to represent the method of
+the given name, if one exists. If not an exception is thrown.
+
+This will additionally search superclasses, and may return a method belonging
+to a parent class.
+
+=head2 direct_methods
+
+   @metamethods = $metaclass->direct_methods
+
+I<Since version 0.57.>
+
+Returns a list of L<Object::Pad::MOP::Method> instances to represent all the
+direct methods of the class. This list may be empty.
+
+=head2 all_methods
+
+   @metamethods = $metaclass->all_methods
+
+I<Since version 0.57.>
+
+Returns a list of L<Object::Pad::MOP::Method> instances to represent all the
+methods of the class, including those inherited from superclasses. This list
+may be empty.
+
 =head2 add_slot
 
    $metaslot = $metaclass->add_slot( $name, %args )
@@ -213,9 +263,13 @@ set (such value may still be C<undef>).
 
 I<Since version 0.46.>
 
-Provides method names for generated reader, writer or lvalue-mutator accessor
-methods, similar to setting them via the C<:reader>, C<:writer> or C<:mutator>
-attributes.
+=item accessor => STRING
+
+I<Since version 0.56.>
+
+Provides method names for generated reader, writer, lvalue-mutator or
+reader+writer accessor methods, similar to setting them via the C<:reader>,
+C<:writer>, C<:mutator> or C<:accessor> attributes.
 
 =item weak => BOOL
 
@@ -246,6 +300,10 @@ slots of the class. This list may be empty.
 
 =cut
 
+*roles = \&direct_roles;
+
+*get_own_method = \&get_direct_method;
+
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
