@@ -1,3 +1,4 @@
+/* vi: set ft=xs : */
 #define PERL_NO_GET_CONTEXT
 
 #include "EXTERN.h"
@@ -30,17 +31,19 @@ struct ClassAttributeRegistration {
   STRLEN permit_hintkeylen;
 
   const struct ClassHookFuncs *funcs;
+  void *funcdata;
 };
 
 static ClassAttributeRegistration *classattrs = NULL;
 
-static void register_class_attribute(const char *name, const struct ClassHookFuncs *funcs)
+static void register_class_attribute(const char *name, const struct ClassHookFuncs *funcs, void *funcdata)
 {
   ClassAttributeRegistration *reg;
   Newx(reg, 1, struct ClassAttributeRegistration);
 
   reg->name = name;
   reg->funcs = funcs;
+  reg->funcdata = funcdata;
 
   if(funcs->permit_hintkey)
     reg->permit_hintkeylen = strlen(funcs->permit_hintkey);
@@ -51,10 +54,32 @@ static void register_class_attribute(const char *name, const struct ClassHookFun
   classattrs = reg;
 }
 
-void ObjectPad_register_class_attribute(pTHX_ const char *name, const struct ClassHookFuncs *funcs)
+struct ClassHookFuncs_v51 {
+  U32 ver;
+  U32 flags;
+  const char *permit_hintkey;
+
+  /* At ABIVERSION 51, callback funcs did not take a 'funcdata' parameter */
+  bool (*apply)(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr);
+  void (*post_add_slot)(pTHX_ ClassMeta *classmeta, SV *hookdata, SlotMeta *slotmeta);
+};
+
+static bool classhook_compat_apply_v51(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
 {
-  if(funcs->ver < 50)
-    croak("Mismatch in third-party class attribute ABI version field: module wants %d, we require >= 50\n",
+  struct ClassHookFuncs_v51 *funcdata = _funcdata;
+  return (*funcdata->apply)(aTHX_ classmeta, value, hookdata_ptr);
+}
+
+static void classhook_compat_post_add_slot_v51(pTHX_ ClassMeta *classmeta, SV *hookdata, void *_funcdata, SlotMeta *slotmeta)
+{
+  struct ClassHookFuncs_v51 *funcdata = _funcdata;
+  (*funcdata->post_add_slot)(aTHX_ classmeta, hookdata, slotmeta);
+}
+
+void ObjectPad_register_class_attribute(pTHX_ const char *name, const struct ClassHookFuncs *funcs, void *funcdata)
+{
+  if(funcs->ver < 51)
+    croak("Mismatch in third-party class attribute ABI version field: module wants %d, we require >= 51\n",
         funcs->ver);
   if(funcs->ver > OBJECTPAD_ABIVERSION)
     croak("Mismatch in third-party class attribute ABI version field: attribute supplies %d, module wants %d\n",
@@ -66,7 +91,25 @@ void ObjectPad_register_class_attribute(pTHX_ const char *name, const struct Cla
   if(!funcs->permit_hintkey)
     croak("Third-party class attributes require a permit hinthash key");
 
-  register_class_attribute(name, funcs);
+  if(funcs->ver < 57) {
+    funcdata = (void *)funcs;
+
+    struct ClassHookFuncs *compatfuncs;
+    Newxz(compatfuncs, 1, struct ClassHookFuncs);
+
+    compatfuncs->ver            = OBJECTPAD_ABIVERSION;
+    compatfuncs->flags          = funcs->flags;
+    compatfuncs->permit_hintkey = funcs->permit_hintkey;
+
+    if(funcs->apply)
+      compatfuncs->apply = &classhook_compat_apply_v51;
+    if(funcs->post_add_slot)
+      compatfuncs->post_add_slot = &classhook_compat_post_add_slot_v51;
+
+    funcs = (const struct ClassHookFuncs *)compatfuncs;
+  }
+
+  register_class_attribute(name, funcs, funcdata);
 }
 
 void ObjectPad_mop_class_apply_attribute(pTHX_ ClassMeta *classmeta, const char *name, SV *value)
@@ -93,16 +136,8 @@ void ObjectPad_mop_class_apply_attribute(pTHX_ ClassMeta *classmeta, const char
     SV *hookdata = value;
 
     if(reg->funcs->apply) {
-      if(reg->funcs->ver >= 51) {
-        if(!(*reg->funcs->apply)(aTHX_ classmeta, value, &hookdata))
-          return;
-      }
-      else {
-        /* ABIVERSION_MINOR 50 apply did not have hookdata_ptr */
-        bool (*apply)(pTHX_ ClassMeta *, SV *) = (void *)(reg->funcs->apply);
-        if(!(*apply)(aTHX_ classmeta, value))
-          return;
-      }
+      if(!(*reg->funcs->apply)(aTHX_ classmeta, value, &hookdata, reg->funcdata))
+        return;
     }
 
     if(!classmeta->hooks)
@@ -112,6 +147,7 @@ void ObjectPad_mop_class_apply_attribute(pTHX_ ClassMeta *classmeta, const char
     Newx(hook, 1, struct ClassHook);
 
     hook->funcs = reg->funcs;
+    hook->funcdata = reg->funcdata;
     hook->hookdata = hookdata;
 
     av_push(classmeta->hooks, (SV *)hook);
@@ -145,12 +181,13 @@ static void S_make_instance_slots(pTHX_ const ClassMeta *classmeta, AV *slotsav,
 
   if(classmeta->start_slotix) {
     /* Superclass actually has some slots */
-    assert(classmeta->supermeta->sealed);
+    assert(classmeta->type == METATYPE_CLASS);
+    assert(classmeta->cls.supermeta->sealed);
 
-    make_instance_slots(classmeta->supermeta, slotsav, 0);
+    make_instance_slots(classmeta->cls.supermeta, slotsav, 0);
   }
 
-  AV *slots = classmeta->slots;
+  AV *slots = classmeta->direct_slots;
   I32 nslots = av_count(slots);
 
   av_extend(slotsav, classmeta->next_slotix - 1 + roleoffset);
@@ -180,18 +217,20 @@ static void S_make_instance_slots(pTHX_ const ClassMeta *classmeta, AV *slotsav,
     }
   }
 
-  AV *roles = classmeta->roles;
-  I32 nroles = av_count(roles);
+  if(classmeta->type == METATYPE_CLASS) {
+    U32 nroles;
+    RoleEmbedding **embeddings = mop_class_get_direct_roles(classmeta, &nroles);
 
-  assert(classmeta->type == METATYPE_CLASS || nroles == 0);
+    assert(classmeta->type == METATYPE_CLASS || nroles == 0);
 
-  for(i = 0; i < nroles; i++) {
-    RoleEmbedding *embedding = (RoleEmbedding *)AvARRAY(roles)[i];
-    ClassMeta *rolemeta = embedding->rolemeta;
+    for(i = 0; i < nroles; i++) {
+      RoleEmbedding *embedding = embeddings[i];
+      ClassMeta *rolemeta = embedding->rolemeta;
 
-    assert(rolemeta->sealed);
+      assert(rolemeta->sealed);
 
-    make_instance_slots(rolemeta, slotsav, embedding->offset);
+      make_instance_slots(rolemeta, slotsav, embedding->offset);
+    }
   }
 }
 
@@ -271,9 +310,25 @@ static CV *S_embed_cv(pTHX_ CV *cv, RoleEmbedding *embedding)
   return embedded_cv;
 }
 
+RoleEmbedding **ObjectPad_mop_class_get_direct_roles(pTHX_ const ClassMeta *meta, U32 *nroles)
+{
+  assert(meta->type == METATYPE_CLASS);
+  AV *roles = meta->cls.direct_roles;
+  *nroles = av_count(roles);
+  return (RoleEmbedding **)AvARRAY(roles);
+}
+
+RoleEmbedding **ObjectPad_mop_class_get_all_roles(pTHX_ const ClassMeta *meta, U32 *nroles)
+{
+  assert(meta->type == METATYPE_CLASS);
+  AV *roles = meta->cls.embedded_roles;
+  *nroles = av_count(roles);
+  return (RoleEmbedding **)AvARRAY(roles);
+}
+
 MethodMeta *ObjectPad_mop_class_add_method(pTHX_ ClassMeta *meta, SV *methodname)
 {
-  AV *methods = meta->methods;
+  AV *methods = meta->direct_methods;
 
   if(meta->sealed)
     croak("Cannot add a new method to an already-sealed class");
@@ -307,8 +362,10 @@ MethodMeta *ObjectPad_mop_class_add_method(pTHX_ ClassMeta *meta, SV *methodname
 
 SlotMeta *ObjectPad_mop_class_add_slot(pTHX_ ClassMeta *meta, SV *slotname)
 {
-  AV *slots = meta->slots;
+  AV *slots = meta->direct_slots;
 
+  if(meta->next_slotix == -1)
+    croak("Cannot add a new slot to a class that is not yet begun");
   if(meta->sealed)
     croak("Cannot add a new slot to an already-sealed class");
 
@@ -347,6 +404,8 @@ SlotMeta *ObjectPad_mop_class_add_slot(pTHX_ ClassMeta *meta, SV *slotname)
 
 void ObjectPad_mop_class_add_BUILD(pTHX_ ClassMeta *meta, CV *cv)
 {
+  if(meta->sealed)
+    croak("Cannot add a BUILD block to an already-sealed class");
   if(meta->strict_params)
     croak("Cannot add a BUILD block to a class with :strict(params)");
 
@@ -358,6 +417,8 @@ void ObjectPad_mop_class_add_BUILD(pTHX_ ClassMeta *meta, CV *cv)
 
 void ObjectPad_mop_class_add_ADJUST(pTHX_ ClassMeta *meta, CV *cv)
 {
+  if(meta->sealed)
+    croak("Cannot add an ADJUST block to an already-sealed class");
   if(!meta->adjustblocks)
     meta->adjustblocks = newAV();
 
@@ -372,6 +433,8 @@ void ObjectPad_mop_class_add_ADJUST(pTHX_ ClassMeta *meta, CV *cv)
 
 void ObjectPad_mop_class_add_ADJUSTPARAMS(pTHX_ ClassMeta *meta, CV *cv)
 {
+  if(meta->sealed)
+    croak("Cannot add an ADJUSTPARAMS block to an already-sealed class");
   if(!meta->adjustblocks)
     meta->adjustblocks = newAV();
 
@@ -386,22 +449,39 @@ void ObjectPad_mop_class_add_ADJUSTPARAMS(pTHX_ ClassMeta *meta, CV *cv)
   av_push(meta->adjustblocks, (SV *)block);
 }
 
-#define mop_class_implements_role(classmeta, rolemeta)  S_mop_class_implements_role(aTHX_ classmeta, rolemeta)
-static bool S_mop_class_implements_role(pTHX_ ClassMeta *classmeta, ClassMeta *rolemeta)
+#define mop_class_implements_role(meta, rolemeta)  S_mop_class_implements_role(aTHX_ meta, rolemeta)
+static bool S_mop_class_implements_role(pTHX_ ClassMeta *meta, ClassMeta *rolemeta)
 {
-  U32 i, n = av_count(classmeta->roles);
-  RoleEmbedding **arr = (RoleEmbedding **)AvARRAY(classmeta->roles);
-  for(i = 0; i < n; i++)
-    if(arr[i]->rolemeta == rolemeta)
-      return true;
+  U32 i, n;
+  switch(meta->type) {
+    case METATYPE_CLASS: {
+      RoleEmbedding **embeddings = mop_class_get_all_roles(meta, &n);
+      for(i = 0; i < n; i++)
+        if(embeddings[i]->rolemeta == rolemeta)
+          return true;
 
-  if(classmeta->supermeta && mop_class_implements_role(classmeta->supermeta, rolemeta))
-    return true;
+      break;
+    }
+
+    case METATYPE_ROLE: {
+      ClassMeta **roles = (ClassMeta **)AvARRAY(meta->role.superroles);
+      U32 n = av_count(meta->role.superroles);
+      /* TODO: this isn't super-efficient in deep cross-linked heirarchies */
+      for(i = 0; i < n; i++) {
+        if(roles[i] == rolemeta)
+          return true;
+        if(mop_class_implements_role(roles[i], rolemeta))
+          return true;
+      }
+      break;
+    }
+  }
 
   return false;
 }
 
-void ObjectPad_mop_class_compose_role(pTHX_ ClassMeta *classmeta, ClassMeta *rolemeta)
+#define embed_role(class, role)  S_embed_role(aTHX_ class, role)
+static RoleEmbedding *S_embed_role(pTHX_ ClassMeta *classmeta, ClassMeta *rolemeta)
 {
   U32 i;
 
@@ -410,9 +490,6 @@ void ObjectPad_mop_class_compose_role(pTHX_ ClassMeta *classmeta, ClassMeta *rol
   if(rolemeta->type != METATYPE_ROLE)
     croak("Can only apply a role to a class");
 
-  if(mop_class_implements_role(classmeta, rolemeta))
-    return;
-
   HV *srcstash = rolemeta->stash;
   HV *dststash = classmeta->stash;
 
@@ -426,7 +503,7 @@ void ObjectPad_mop_class_compose_role(pTHX_ ClassMeta *classmeta, ClassMeta *rol
   embedding->classmeta   = classmeta;
   embedding->offset      = -1;
 
-  av_push(classmeta->roles, (SV *)embedding);
+  av_push(classmeta->cls.embedded_roles, (SV *)embedding);
   hv_store_ent(rolemeta->role.applied_classes, classmeta->name, (SV *)embedding, 0);
 
   U32 nbuilds = rolemeta->buildblocks ? av_count(rolemeta->buildblocks) : 0;
@@ -456,9 +533,9 @@ void ObjectPad_mop_class_compose_role(pTHX_ ClassMeta *classmeta, ClassMeta *rol
   if(rolemeta->has_adjustparams)
     classmeta->has_adjustparams = true;
 
-  U32 nmethods = av_count(rolemeta->methods);
+  U32 nmethods = av_count(rolemeta->direct_methods);
   for(i = 0; i < nmethods; i++) {
-    MethodMeta *methodmeta = (MethodMeta *)AvARRAY(rolemeta->methods)[i];
+    MethodMeta *methodmeta = (MethodMeta *)AvARRAY(rolemeta->direct_methods)[i];
     SV *mname = methodmeta->name;
 
     HE *he = hv_fetch_ent(srcstash, mname, 0, 0);
@@ -488,6 +565,67 @@ void ObjectPad_mop_class_compose_role(pTHX_ ClassMeta *classmeta, ClassMeta *rol
   for(i = 0; i < nmethods; i++) {
     av_push(classmeta->requiremethods, SvREFCNT_inc(AvARRAY(rolemeta->requiremethods)[i]));
   }
+
+  return embedding;
+}
+
+void ObjectPad_mop_class_add_role(pTHX_ ClassMeta *dstmeta, ClassMeta *rolemeta)
+{
+  if(dstmeta->sealed)
+    croak("Cannot add a role to an already-sealed class");
+  /* Can't currently do this as it breaks t/77mop-create-role.t
+  if(!rolemeta->sealed)
+    croak("Cannot add a role that is not yet sealed");
+   */
+
+  if(mop_class_implements_role(dstmeta, rolemeta))
+    return;
+
+  switch(dstmeta->type) {
+    case METATYPE_CLASS: {
+      U32 nroles;
+      if((nroles = av_count(rolemeta->role.superroles)) > 0) {
+        ClassMeta **roles = (ClassMeta **)AvARRAY(rolemeta->role.superroles);
+        U32 i;
+        for(i = 0; i < nroles; i++)
+          mop_class_add_role(dstmeta, roles[i]);
+      }
+
+      RoleEmbedding *embedding = embed_role(dstmeta, rolemeta);
+      av_push(dstmeta->cls.direct_roles, (SV *)embedding);
+      return;
+    }
+
+    case METATYPE_ROLE:
+      av_push(dstmeta->role.superroles, (SV *)rolemeta);
+      return;
+  }
+}
+
+void ObjectPad_mop_class_load_and_add_role(pTHX_ ClassMeta *meta, SV *rolename, SV *rolever)
+{
+  HV *rolestash = gv_stashsv(rolename, 0);
+  if(!rolestash || !hv_fetchs(rolestash, "META", 0)) {
+    /* Try to`require` the module then attempt a second time */
+    load_module(PERL_LOADMOD_NOIMPORT, newSVsv(rolename), NULL, NULL);
+    rolestash = gv_stashsv(rolename, 0);
+  }
+
+  if(!rolestash)
+    croak("Role %" SVf " does not exist", SVfARG(rolename));
+
+  if(rolever && SvOK(rolever))
+    ensure_module_version(rolename, rolever);
+
+  GV **metagvp = (GV **)hv_fetchs(rolestash, "META", 0);
+  ClassMeta *rolemeta = NULL;
+  if(metagvp)
+    rolemeta = NUM2PTR(ClassMeta *, SvUV(SvRV(GvSV(*metagvp))));
+
+  if(!rolemeta || rolemeta->type != METATYPE_ROLE)
+    croak("%" SVf " is not a role", SVfARG(rolename));
+
+  mop_class_add_role(meta, rolemeta);
 }
 
 #define embed_slothook(roleh, offset)  S_embed_slothook(aTHX_ roleh, offset)
@@ -515,6 +653,7 @@ static void S_mop_class_apply_role(pTHX_ RoleEmbedding *embedding)
   if(rolemeta->type != METATYPE_ROLE)
     croak("Can only apply a role to a class");
 
+  assert(embedding->offset == -1);
   embedding->offset = classmeta->next_slotix;
 
   if(rolemeta->parammap) {
@@ -573,20 +712,18 @@ static void S_mop_class_apply_role(pTHX_ RoleEmbedding *embedding)
     }
   }
 
-  classmeta->next_slotix += av_count(rolemeta->slots);
+  classmeta->next_slotix += av_count(rolemeta->direct_slots);
 
   /* TODO: Run an APPLY block if the role has one */
 }
 
-static void S_apply_roles(pTHX_ ClassMeta *dst, ClassMeta *src)
+static void S_apply_roles(pTHX_ ClassMeta *dstmeta, ClassMeta *srcmeta)
 {
-  U32 nroles = av_count(src->roles);
-  RoleEmbedding **arr = (RoleEmbedding **)AvARRAY(src->roles);
+  U32 nroles;
+  RoleEmbedding **arr = mop_class_get_direct_roles(srcmeta, &nroles);
   U32 i;
   for(i = 0; i < nroles; i++) {
     mop_class_apply_role(arr[i]);
-
-    /* TODO: Consider how we recurse into roles */
   }
 }
 
@@ -675,7 +812,7 @@ static void S_generate_initslots_method(pTHX_ ClassMeta *meta)
    * we must be a subclass
    */
   if(meta->start_slotix) {
-    struct ClassMeta *supermeta = meta->supermeta;
+    struct ClassMeta *supermeta = meta->cls.supermeta;
 
     assert(supermeta->sealed);
     assert(supermeta->initslots);
@@ -698,31 +835,32 @@ static void S_generate_initslots_method(pTHX_ ClassMeta *meta)
       op_convert_list(OP_ENTERSUB, OPf_WANT_VOID|OPf_STACKED, op));
   }
 
-  AV *slots = meta->slots;
+  AV *slots = meta->direct_slots;
   I32 nslots = av_count(slots);
 
   {
-    CopLINE_set(PL_curcop, __LINE__);
-
     for(i = 0; i < nslots; i++) {
       SlotMeta *slotmeta = (SlotMeta *)AvARRAY(slots)[i];
       char sigil = SvPV_nolen(slotmeta->name)[0];
       OP *op = NULL;
+      SV *defaultsv;
 
       switch(sigil) {
         case '$':
         {
+          CopLINE_set(PL_curcop, __LINE__);
+
           OP *valueop = NULL;
 
           if(slotmeta->defaultexpr) {
             valueop = slotmeta->defaultexpr;
           }
-          else if(slotmeta->defaultsv) {
+          else if((defaultsv = mop_slot_get_default_sv(slotmeta))) {
             /* An OP_CONST whose op_type is OP_CUSTOM.
              * This way we avoid the opchecker and finalizer doing bad things
              * to our defaultsv SV by setting it SvREADONLY_on()
              */
-            valueop = newSVOP_CUSTOM(PL_ppaddr[OP_CONST], 0, slotmeta->defaultsv);
+            valueop = newSVOP_CUSTOM(PL_ppaddr[OP_CONST], 0, defaultsv);
           }
 
           if(slotmeta->paramname) {
@@ -755,7 +893,7 @@ static void S_generate_initslots_method(pTHX_ ClassMeta *meta)
             op = newBINOP(OP_SASSIGN, 0,
               valueop,
               /* $slots[$idx] */
-              newAELEMOP(0,
+              newAELEMOP(OPf_MOD,
                 newPADxVOP(OP_PADAV, OPf_MOD|OPf_REF, PADIX_SLOTS),
                 slotmeta->slotix));
           break;
@@ -763,15 +901,17 @@ static void S_generate_initslots_method(pTHX_ ClassMeta *meta)
         case '@':
         case '%':
         {
+          CopLINE_set(PL_curcop, __LINE__);
+
           OP *valueop = NULL;
           U16 coerceop = (sigil == '%') ? OP_RV2HV : OP_RV2AV;
 
           if(slotmeta->defaultexpr) {
             valueop = slotmeta->defaultexpr;
           }
-          else if(slotmeta->defaultsv) {
+          else if((defaultsv = mop_slot_get_default_sv(slotmeta))) {
             valueop = newUNOP(coerceop, 0,
-                newSVOP_CUSTOM(PL_ppaddr[OP_CONST], 0, slotmeta->defaultsv));
+                newSVOP_CUSTOM(PL_ppaddr[OP_CONST], 0, defaultsv));
           }
 
           if(valueop) {
@@ -803,34 +943,36 @@ static void S_generate_initslots_method(pTHX_ ClassMeta *meta)
     }
   }
 
-  AV *roles = meta->roles;
-  I32 nroles = av_count(roles);
+  if(meta->type == METATYPE_CLASS) {
+    U32 nroles;
+    RoleEmbedding **embeddings = mop_class_get_direct_roles(meta, &nroles);
 
-  for(i = 0; i < nroles; i++) {
-    RoleEmbedding *embedding = (RoleEmbedding *)AvARRAY(roles)[i];
-    ClassMeta *rolemeta = embedding->rolemeta;
+    for(i = 0; i < nroles; i++) {
+      RoleEmbedding *embedding = embeddings[i];
+      ClassMeta *rolemeta = embedding->rolemeta;
 
-    if(!rolemeta->sealed)
-      mop_class_seal(rolemeta);
+      if(!rolemeta->sealed)
+        mop_class_seal(rolemeta);
 
-    assert(rolemeta->sealed);
-    assert(rolemeta->initslots);
+      assert(rolemeta->sealed);
+      assert(rolemeta->initslots);
 
-    CopLINE_set(PL_curcop, __LINE__);
+      CopLINE_set(PL_curcop, __LINE__);
 
-    ops = op_append_list(OP_LINESEQ, ops,
-      newSTATEOP(0, NULL, NULL));
+      ops = op_append_list(OP_LINESEQ, ops,
+        newSTATEOP(0, NULL, NULL));
 
-    OP *op = NULL;
-    op = op_append_list(OP_LIST, op,
-      newPADxVOP(OP_PADSV, 0, PADIX_SELF));
-    op = op_append_list(OP_LIST, op,
-      newPADxVOP(OP_PADHV, OPf_REF, PADIX_INITSLOTS_PARAMS));
-    op = op_append_list(OP_LIST, op,
-      newSVOP(OP_CONST, 0, (SV *)embed_cv(rolemeta->initslots, embedding)));
+      OP *op = NULL;
+      op = op_append_list(OP_LIST, op,
+        newPADxVOP(OP_PADSV, 0, PADIX_SELF));
+      op = op_append_list(OP_LIST, op,
+        newPADxVOP(OP_PADHV, OPf_REF, PADIX_INITSLOTS_PARAMS));
+      op = op_append_list(OP_LIST, op,
+        newSVOP(OP_CONST, 0, (SV *)embed_cv(rolemeta->initslots, embedding)));
 
-    ops = op_append_list(OP_LINESEQ, ops,
-      op_convert_list(OP_ENTERSUB, OPf_WANT_VOID|OPf_STACKED, op));
+      ops = op_append_list(OP_LINESEQ, ops,
+        op_convert_list(OP_ENTERSUB, OPf_WANT_VOID|OPf_STACKED, op));
+    }
   }
 
   SvREFCNT_inc(PL_compcv);
@@ -853,17 +995,20 @@ void ObjectPad_mop_class_seal(pTHX_ ClassMeta *meta)
   if(meta->sealed) /* idempotent */
     return;
 
-  if(meta->supermeta && !meta->supermeta->sealed) {
+  if(meta->type == METATYPE_CLASS &&
+      meta->cls.supermeta && !meta->cls.supermeta->sealed) {
     /* Must defer sealing until superclass is sealed first
      * (RT133190)
      */
-    if(!meta->supermeta->pending_submeta)
-      meta->supermeta->pending_submeta = newAV();
-    av_push(meta->supermeta->pending_submeta, (SV *)meta);
+    ClassMeta *supermeta = meta->cls.supermeta;
+    if(!supermeta->pending_submeta)
+      supermeta->pending_submeta = newAV();
+    av_push(supermeta->pending_submeta, (SV *)meta);
     return;
   }
 
-  S_apply_roles(aTHX_ meta, meta);
+  if(meta->type == METATYPE_CLASS)
+    S_apply_roles(aTHX_ meta, meta);
 
   if(meta->type == METATYPE_CLASS) {
     U32 nmethods = av_count(meta->requiremethods);
@@ -886,8 +1031,8 @@ void ObjectPad_mop_class_seal(pTHX_ ClassMeta *meta)
 
   {
     U32 slotix;
-    for(slotix = 0; slotix < av_count(meta->slots); slotix++) {
-      SlotMeta *slotmeta = (SlotMeta *)AvARRAY(meta->slots)[slotix];
+    for(slotix = 0; slotix < av_count(meta->direct_slots); slotix++) {
+      SlotMeta *slotmeta = (SlotMeta *)AvARRAY(meta->direct_slots)[slotix];
 
       U32 hooki;
       for(hooki = 0; slotmeta->hooks && hooki < av_count(slotmeta->hooks); hooki++) {
@@ -903,6 +1048,7 @@ void ObjectPad_mop_class_seal(pTHX_ ClassMeta *meta)
           fasth->slotix   = slotix;
           fasth->slotmeta = slotmeta;
           fasth->funcs    = h->funcs;
+          fasth->funcdata = h->funcdata;
           fasth->hookdata = h->hookdata;
 
           av_push(meta->slothooks_postslots, (SV *)fasth);
@@ -918,6 +1064,7 @@ void ObjectPad_mop_class_seal(pTHX_ ClassMeta *meta)
           fasth->slotix   = slotix;
           fasth->slotmeta = slotmeta;
           fasth->funcs    = h->funcs;
+          fasth->funcdata = h->funcdata;
           fasth->hookdata = h->hookdata;
 
           av_push(meta->slothooks_construct, (SV *)fasth);
@@ -970,10 +1117,10 @@ XS_INTERNAL(injected_constructor)
 
   {
     /* @args = $class->BUILDARGS(@_) */
-    CopLINE_set(PL_curcop, __LINE__);
-
     ENTER;
     SAVETMPS;
+    SAVEVPTR(PL_curcop);
+    PL_curcop = prevcop;
 
     /* Splice in an extra copy of `class` so we get one there for the foreign
      * constructor */
@@ -1134,7 +1281,7 @@ XS_INTERNAL(injected_constructor)
       struct SlotHook *h = (struct SlotHook *)AvARRAY(slothooks)[i];
       SLOTOFFSET slotix = h->slotix;
 
-      (*h->funcs->post_initslot)(aTHX_ h->slotmeta, h->hookdata, slotsv[slotix]);
+      (*h->funcs->post_initslot)(aTHX_ h->slotmeta, h->hookdata, h->funcdata, slotsv[slotix]);
     }
   }
 
@@ -1266,7 +1413,7 @@ XS_INTERNAL(injected_constructor)
       struct SlotHook *h = (struct SlotHook *)AvARRAY(slothooks)[i];
       SLOTOFFSET slotix = h->slotix;
 
-      (*h->funcs->post_construct)(aTHX_ h->slotmeta, h->hookdata, slotsv[slotix]);
+      (*h->funcs->post_construct)(aTHX_ h->slotmeta, h->hookdata, h->funcdata, slotsv[slotix]);
     }
   }
 
@@ -1287,8 +1434,8 @@ XS_INTERNAL(injected_DOES)
   CV *cv_does = NULL;
 
   while(meta != NULL) {
-    AV *roles = meta->roles;
-    I32 nroles = av_count(roles);
+    AV *roles = meta->type == METATYPE_CLASS ? meta->cls.direct_roles : NULL;
+    I32 nroles = roles ? av_count(roles) : 0;
 
     if(!cv_does && meta->cls.foreign_does)
       cv_does = meta->cls.foreign_does;
@@ -1305,7 +1452,7 @@ XS_INTERNAL(injected_DOES)
       }
     }
 
-    meta = meta->supermeta;
+    meta = meta->type == METATYPE_CLASS ? meta->cls.supermeta : NULL;
   }
 
   if (cv_does) {
@@ -1345,12 +1492,9 @@ XS_INTERNAL(injected_DOES)
   XSRETURN_NO;
 }
 
-ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name, SV *superclassname)
+ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name)
 {
-  assert(
-      type == METATYPE_CLASS ||
-      (type == METATYPE_ROLE && !superclassname)
-  );
+  assert(type == METATYPE_CLASS || type == METATYPE_ROLE);
 
   ClassMeta *meta;
   Newx(meta, 1, ClassMeta);
@@ -1364,16 +1508,16 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name, SV *su
   meta->role_is_invokable = false;
   meta->strict_params = false;
   meta->has_adjustparams = false;
+  meta->has_superclass = false;
   meta->start_slotix = 0;
+  meta->next_slotix = -1;
   meta->hooks   = NULL;
-  meta->slots   = newAV();
-  meta->methods = newAV();
+  meta->direct_slots = newAV();
+  meta->direct_methods = newAV();
   meta->parammap = NULL;
   meta->requiremethods = newAV();
   meta->repr   = REPR_AUTOSELECT;
-  meta->supermeta = NULL;
   meta->pending_submeta = NULL;
-  meta->roles = newAV();
   meta->buildblocks = NULL;
   meta->adjustblocks = NULL;
   meta->initslots = NULL;
@@ -1383,11 +1527,15 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name, SV *su
 
   switch(type) {
     case METATYPE_CLASS:
+      meta->cls.supermeta = NULL;
       meta->cls.foreign_new = NULL;
       meta->cls.foreign_does = NULL;
+      meta->cls.direct_roles = newAV();
+      meta->cls.embedded_roles = newAV();
       break;
 
     case METATYPE_ROLE:
+      meta->role.superroles = newAV();
       meta->role.applied_classes = newHV();
       break;
   }
@@ -1445,123 +1593,6 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name, SV *su
 
   meta->methodscope = NULL;
 
-  AV *isa;
-  {
-    SV *isaname = newSVpvf("%" SVf "::ISA", name);
-    SAVEFREESV(isaname);
-
-    isa = get_av(SvPV_nolen(isaname), GV_ADD | (SvFLAGS(isaname) & SVf_UTF8));
-  }
-
-  if(superclassname && SvOK(superclassname)) {
-    assert(type == METATYPE_CLASS);
-
-    av_push(isa, SvREFCNT_inc(superclassname));
-
-    ClassMeta *supermeta = NULL;
-
-    HV *superstash = gv_stashsv(superclassname, 0);
-    GV **metagvp = (GV **)hv_fetchs(superstash, "META", 0);
-    if(metagvp)
-      supermeta = NUM2PTR(ClassMeta *, SvUV(SvRV(GvSV(*metagvp))));
-
-    if(supermeta) {
-      /* A subclass of an Object::Pad class */
-      if(supermeta->type != METATYPE_CLASS)
-        croak("%" SVf " is not a class", SVfARG(superclassname));
-
-      meta->start_slotix = supermeta->next_slotix;
-      meta->repr = supermeta->repr;
-      meta->cls.foreign_new = supermeta->cls.foreign_new;
-
-      if(supermeta->buildblocks) {
-        if(!meta->buildblocks)
-          meta->buildblocks = newAV();
-
-        av_push_from_av_noinc(meta->buildblocks, supermeta->buildblocks);
-      }
-
-      if(supermeta->adjustblocks) {
-        if(!meta->adjustblocks)
-          meta->adjustblocks = newAV();
-
-        av_push_from_av_noinc(meta->adjustblocks, supermeta->adjustblocks);
-      }
-
-      if(supermeta->slothooks_postslots) {
-        if(!meta->slothooks_postslots)
-          meta->slothooks_postslots = newAV();
-
-        av_push_from_av_noinc(meta->slothooks_postslots, supermeta->slothooks_postslots);
-      }
-
-      if(supermeta->slothooks_construct) {
-        if(!meta->slothooks_construct)
-          meta->slothooks_construct = newAV();
-
-        av_push_from_av_noinc(meta->slothooks_construct, supermeta->slothooks_construct);
-      }
-
-      if(supermeta->parammap) {
-        HV *old = supermeta->parammap;
-        HV *new = meta->parammap = newHV();
-
-        hv_iterinit(old);
-
-        HE *iter;
-        while((iter = hv_iternext(old))) {
-          STRLEN klen = HeKLEN(iter);
-          /* Don't SvREFCNT_inc() the values because they aren't really SV *s */
-          /* Subclasses *DIRECTLY SHARE* their param metas because the
-           * information in them is directly compatible
-           */
-          if(klen < 0)
-            hv_store_ent(new, HeSVKEY(iter), HeVAL(iter), HeHASH(iter));
-          else
-            hv_store(new, HeKEY(iter), klen, HeVAL(iter), HeHASH(iter));
-        }
-      }
-
-      if(supermeta->has_adjustparams)
-        meta->has_adjustparams = true;
-
-      /* Mark that this class is applied to every role our supermetas have */
-      ClassMeta *c;
-      for(c = supermeta; c; c = c->supermeta) {
-        AV *roles = c->roles;
-        U32 nroles = av_count(roles);
-        U32 i;
-        for(i = 0; i < nroles; i++) {
-          RoleEmbedding *embedding = (RoleEmbedding *)AvARRAY(roles)[i];
-          ClassMeta *rolemeta = embedding->rolemeta;
-
-          hv_store_ent(rolemeta->role.applied_classes, meta->name, (SV *)embedding, 0);
-        }
-      }
-    }
-    else {
-      /* A subclass of a foreign class */
-      meta->cls.foreign_new = fetch_superclass_method_pv(meta->stash, "new", 3, -1);
-      if(!meta->cls.foreign_new)
-        croak("Unable to find SUPER::new for %" SVf, superclassname);
-
-      meta->cls.foreign_does = fetch_superclass_method_pv(meta->stash, "DOES", 4, -1);
-
-      av_push(isa, newSVpvs("Object::Pad::UNIVERSAL"));
-    }
-
-    meta->supermeta = supermeta;
-  }
-  else {
-    /* A base class */
-    av_push(isa, newSVpvs("Object::Pad::UNIVERSAL"));
-  }
-
-  if(meta->repr == REPR_AUTOSELECT && !meta->cls.foreign_new)
-    meta->repr = REPR_NATIVE;
-
-  meta->next_slotix = meta->start_slotix;
-
   {
     /* Inject the constructor */
     SV *newname = newSVpvf("%" SVf "::new", name);
@@ -1593,13 +1624,247 @@ ClassMeta *ObjectPad_mop_create_class(pTHX_ enum MetaType type, SV *name, SV *su
   return meta;
 }
 
+void ObjectPad_mop_class_set_superclass(pTHX_ ClassMeta *meta, SV *superclassname)
+{
+  assert(meta->type == METATYPE_CLASS);
+
+  if(meta->has_superclass)
+    croak("Class already has a superclass, cannot add another");
+
+  AV *isa;
+  {
+    SV *isaname = newSVpvf("%" SVf "::ISA", meta->name);
+    SAVEFREESV(isaname);
+
+    isa = get_av(SvPV_nolen(isaname), GV_ADD | (SvFLAGS(isaname) & SVf_UTF8));
+  }
+
+  av_push(isa, SvREFCNT_inc(superclassname));
+
+  ClassMeta *supermeta = NULL;
+
+  HV *superstash = gv_stashsv(superclassname, 0);
+  GV **metagvp = (GV **)hv_fetchs(superstash, "META", 0);
+  if(metagvp)
+    supermeta = NUM2PTR(ClassMeta *, SvUV(SvRV(GvSV(*metagvp))));
+
+  if(supermeta) {
+    /* A subclass of an Object::Pad class */
+    if(supermeta->type != METATYPE_CLASS)
+      croak("%" SVf " is not a class", SVfARG(superclassname));
+
+    meta->start_slotix = supermeta->next_slotix;
+    meta->repr = supermeta->repr;
+    meta->cls.foreign_new = supermeta->cls.foreign_new;
+
+    if(supermeta->buildblocks) {
+      if(!meta->buildblocks)
+        meta->buildblocks = newAV();
+
+      av_push_from_av_noinc(meta->buildblocks, supermeta->buildblocks);
+    }
+
+    if(supermeta->adjustblocks) {
+      if(!meta->adjustblocks)
+        meta->adjustblocks = newAV();
+
+      av_push_from_av_noinc(meta->adjustblocks, supermeta->adjustblocks);
+    }
+
+    if(supermeta->slothooks_postslots) {
+      if(!meta->slothooks_postslots)
+        meta->slothooks_postslots = newAV();
+
+      av_push_from_av_noinc(meta->slothooks_postslots, supermeta->slothooks_postslots);
+    }
+
+    if(supermeta->slothooks_construct) {
+      if(!meta->slothooks_construct)
+        meta->slothooks_construct = newAV();
+
+      av_push_from_av_noinc(meta->slothooks_construct, supermeta->slothooks_construct);
+    }
+
+    if(supermeta->parammap) {
+      HV *old = supermeta->parammap;
+      HV *new = meta->parammap = newHV();
+
+      hv_iterinit(old);
+
+      HE *iter;
+      while((iter = hv_iternext(old))) {
+        STRLEN klen = HeKLEN(iter);
+        /* Don't SvREFCNT_inc() the values because they aren't really SV *s */
+        /* Subclasses *DIRECTLY SHARE* their param metas because the
+         * information in them is directly compatible
+         */
+        if(klen < 0)
+          hv_store_ent(new, HeSVKEY(iter), HeVAL(iter), HeHASH(iter));
+        else
+          hv_store(new, HeKEY(iter), klen, HeVAL(iter), HeHASH(iter));
+      }
+    }
+
+    if(supermeta->has_adjustparams)
+      meta->has_adjustparams = true;
+
+    U32 nroles;
+    RoleEmbedding **embeddings = mop_class_get_all_roles(supermeta, &nroles);
+    if(nroles) {
+      U32 i;
+      for(i = 0; i < nroles; i++) {
+        RoleEmbedding *embedding = embeddings[i];
+        ClassMeta *rolemeta = embedding->rolemeta;
+
+        av_push(meta->cls.embedded_roles, (SV *)embedding);
+        hv_store_ent(rolemeta->role.applied_classes, meta->name, (SV *)embedding, 0);
+      }
+    }
+  }
+  else {
+    /* A subclass of a foreign class */
+    meta->cls.foreign_new = fetch_superclass_method_pv(meta->stash, "new", 3, -1);
+    if(!meta->cls.foreign_new)
+      croak("Unable to find SUPER::new for %" SVf, superclassname);
+
+    meta->cls.foreign_does = fetch_superclass_method_pv(meta->stash, "DOES", 4, -1);
+
+    av_push(isa, newSVpvs("Object::Pad::UNIVERSAL"));
+  }
+
+  meta->has_superclass = true;
+  meta->cls.supermeta = supermeta;
+}
+
+void ObjectPad_mop_class_begin(pTHX_ ClassMeta *meta)
+{
+  SV *isaname = newSVpvf("%" SVf "::ISA", meta->name);
+  SAVEFREESV(isaname);
+
+  AV *isa = get_av(SvPV_nolen(isaname), GV_ADD | (SvFLAGS(isaname) & SVf_UTF8));
+  if(!av_count(isa))
+    av_push(isa, newSVpvs("Object::Pad::UNIVERSAL"));
+
+  if(meta->type == METATYPE_CLASS &&
+      meta->repr == REPR_AUTOSELECT && !meta->cls.foreign_new)
+    meta->repr = REPR_NATIVE;
+
+  meta->next_slotix = meta->start_slotix;
+}
+
 /*******************
  * Attribute hooks *
  *******************/
 
+#ifndef isSPACE_utf8_safe
+   /* this isn't really safe but it's the best we can do */
+#  define isSPACE_utf8_safe(p, e)  (PERL_UNUSED_ARG(e), isSPACE_utf8(p))
+#endif
+
+#define split_package_ver(value, pkgname, pkgversion)  S_split_package_ver(aTHX_ value, pkgname, pkgversion)
+static const char *S_split_package_ver(pTHX_ SV *value, SV *pkgname, SV *pkgversion)
+{
+  const char *start = SvPVX(value), *p = start, *end = start + SvCUR(value);
+
+  while(*p && !isSPACE_utf8_safe(p, end))
+    p += UTF8SKIP(p);
+
+  sv_setpvn(pkgname, start, p - start);
+  if(SvUTF8(value))
+    SvUTF8_on(pkgname);
+
+  while(*p && isSPACE_utf8_safe(p, end))
+    p += UTF8SKIP(p);
+
+  if(*p) {
+    /* scan_version() gets upset about trailing content. We need to extract
+     * exactly what it wants
+     */
+    start = p;
+    if(*p == 'v')
+      p++;
+    while(*p && strchr("0123456789._", *p))
+      p++;
+    SV *tmpsv = newSVpvn(start, p - start);
+    SAVEFREESV(tmpsv);
+
+    scan_version(SvPVX(tmpsv), pkgversion, FALSE);
+  }
+
+  while(*p && isSPACE_utf8_safe(p, end))
+    p += UTF8SKIP(p);
+
+  return p;
+}
+
+/* :isa */
+
+static bool classhook_isa_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
+{
+  SV *superclassname = newSV(0), *superclassver = newSV(0);
+  SAVEFREESV(superclassname);
+  SAVEFREESV(superclassver);
+
+  const char *end = split_package_ver(value, superclassname, superclassver);
+
+  if(*end)
+    croak("Unexpected characters while parsing :isa() attribute: %s", end);
+
+  if(classmeta->type != METATYPE_CLASS)
+    croak("Only a class may extend another");
+
+  HV *superstash = gv_stashsv(superclassname, 0);
+  if(!superstash || !hv_fetchs(superstash, "new", 0)) {
+    /* Try to `require` the module then attempt a second time */
+    /* load_module() will modify the name argument and take ownership of it */
+    load_module(PERL_LOADMOD_NOIMPORT, newSVsv(superclassname), NULL, NULL);
+    superstash = gv_stashsv(superclassname, 0);
+  }
+
+  if(!superstash)
+    croak("Superclass %" SVf " does not exist", superclassname);
+
+  if(superclassver && SvOK(superclassver))
+    ensure_module_version(superclassname, superclassver);
+
+  mop_class_set_superclass(classmeta, superclassname);
+
+  return FALSE;
+}
+
+static const struct ClassHookFuncs classhooks_isa = {
+  .ver   = OBJECTPAD_ABIVERSION,
+  .flags = OBJECTPAD_FLAG_ATTR_MUST_VALUE,
+  .apply = &classhook_isa_apply,
+};
+
+/* :does */
+
+static bool classhook_does_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
+{
+  SV *rolename = newSV(0), *rolever = newSV(0);
+  SAVEFREESV(rolename);
+  SAVEFREESV(rolever);
+
+  const char *end = split_package_ver(value, rolename, rolever);
+
+  if(*end)
+    croak("Unexpected characters while parsing :does() attribute: %s", end);
+
+  mop_class_load_and_add_role(classmeta, rolename, rolever);
+
+  return FALSE;
+}
+
+static const struct ClassHookFuncs classhooks_does = {
+  .ver   = OBJECTPAD_ABIVERSION,
+  .flags = OBJECTPAD_FLAG_ATTR_MUST_VALUE,
+  .apply = &classhook_does_apply,
+};
+
 /* :repr */
 
-static bool classhook_repr_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr)
+static bool classhook_repr_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
 {
   char *val = SvPV_nolen(value); /* all comparisons are ASCII */
 
@@ -1631,7 +1896,7 @@ static const struct ClassHookFuncs classhooks_repr = {
 
 /* :compat */
 
-static bool classhook_compat_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr)
+static bool classhook_compat_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
 {
   if(strEQ(SvPV_nolen(value), "invokable")) {
     if(classmeta->type != METATYPE_ROLE)
@@ -1653,7 +1918,7 @@ static const struct ClassHookFuncs classhooks_compat = {
 
 /* :strict */
 
-static bool classhook_strict_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr)
+static bool classhook_strict_apply(pTHX_ ClassMeta *classmeta, SV *value, SV **hookdata_ptr, void *_funcdata)
 {
   if(strEQ(SvPV_nolen(value), "params"))
     classmeta->strict_params = TRUE;
@@ -1671,7 +1936,9 @@ static const struct ClassHookFuncs classhooks_strict = {
 
 void ObjectPad__boot_classes(void)
 {
-  register_class_attribute("repr", &classhooks_repr);
-  register_class_attribute("compat", &classhooks_compat);
-  register_class_attribute("strict", &classhooks_strict);
+  register_class_attribute("isa",    &classhooks_isa,    NULL);
+  register_class_attribute("does",   &classhooks_does,   NULL);
+  register_class_attribute("repr",   &classhooks_repr,   NULL);
+  register_class_attribute("compat", &classhooks_compat, NULL);
+  register_class_attribute("strict", &classhooks_strict, NULL);
 }
